#include <FK/FK.h>

const double	BUILDWIDTH	= 25.0;		// ̊{P
const double	SPEED		= 2.0;		// Ԃ̃Xs[h
const double	CIRCUITX	= 150.0;	// R[X X 
const double	CIRCUITY	= 250.0;	// R[X Y 
const double	EPS			= 0.001;	// 덷pl

class Car {
private:
	fk_Model	carModel;		// ԑS̃f
	fk_Model	bodyModel;		// ԑ̃f
	fk_Model	tireModel[4];	// e^Cf
	fk_Model	driverModel[2];	// ^]҃f
	fk_Model	birdModel;		// Վ_f

	fk_Block	body;			// ԑ̌`
	fk_Circle	tire;			// ^C`
	fk_Sphere	driver;			// ^]Ҍ`

public:
	Car(void) {		// RXgN^
		init();
	}

	void		init(void);
	void		entryScene(fk_Scene *, bool);
	fk_Vector	getCarPosition(void);
	fk_Model *	getBirdModel(void);
	void		forward(void);
	void		rotate(fk_Model *, fk_Vector, fk_Vector);
};

class World {
private:
	fk_Model	buildModel[6], groundModel, lightModel[2];
	fk_Block	buildShape, groundShape;
	fk_Light	lightShape;

	void		defLight(int, double, double, double);
	void		makeBuild(int, double, double, double, fk_Material *);

public:
	World(void) {
		init();
	}

	void		init(void);
	void		entryScene(fk_Scene *, bool);
};

int main(int, char *[])
{
	Car				carObj;
	World			worldObj;
	Fl_Window		mainWindow(940, 320, "FK TEST");
	fk_Scene		carViewScene, buildViewScene, birdViewScene;
	fk_Window		carViewWindow(10, 10, 300, 300);
	fk_Window		buildViewWindow(320, 10, 300, 300);
	fk_Window		birdViewWindow(630, 10, 300, 300);
	fk_Color		bgColor(0.2, 0.7, 1.0);

	fk_Model		buildViewModel, birdViewModel;

	mainWindow.end();
	fk_Material::initDefault();

	carObj.init();
	worldObj.init();


	// eEBhEɃobNOEhJ[ݒ
	carViewScene.setBGColor(bgColor);
	buildViewScene.setBGColor(bgColor);
	birdViewScene.setBGColor(bgColor);


	// effBXvCXgɓo^
	worldObj.entryScene(&carViewScene, false);
	worldObj.entryScene(&buildViewScene, true);
	worldObj.entryScene(&birdViewScene, false);

	carObj.entryScene(&carViewScene, false);
	carObj.entryScene(&buildViewScene, true);
	carObj.entryScene(&birdViewScene, true);

	// EBhE̎_ݒ
	buildViewModel.glMoveTo(-250.0, 100.0, 100.0);
	buildViewModel.glFocus(carObj.getCarPosition());
	buildViewModel.glUpvec(0.0, 0.0, 1.0);
	buildViewScene.entryCamera(&buildViewModel);

	// ՃEBhE̎_ݒ
	birdViewScene.entryCamera(carObj.getBirdModel());

	// EBhEփfBXvCXgo^
	carViewWindow.setScene(&carViewScene);
	buildViewWindow.setScene(&buildViewScene);
	birdViewWindow.setScene(&birdViewScene);

	mainWindow.show();
	carViewWindow.show();
	buildViewWindow.show();
	birdViewWindow.show();

	while(true) {

		if(mainWindow.visible() == 0) {
			if(Fl::wait() == 0) {
				break;
			} else {
				continue;
			}
		}

		if(carViewWindow.drawWindow() == 0 ||
		   buildViewWindow.drawWindow() == 0 ||
		   birdViewWindow.drawWindow() == 0) break;

		if(Fl::check() == 0) break;

		if(carViewWindow.winOpenStatus() == false ||
		   buildViewWindow.winOpenStatus() == false ||
		   birdViewWindow.winOpenStatus() == false) {
			continue;
		}

		carObj.forward();
		buildViewModel.glFocus(carObj.getCarPosition());
		buildViewModel.glUpvec(0.0, 0.0, 1.0);
	}

	return 0;
}

void Car::init(void)
{
	int		i;

	body.setSize(7.0, 6.0, 20.0);
	tire.setRadius(2.0);
	tire.setDivide(2);
	driver.setRadius(2.0);
	driver.setDivide(2);

	bodyModel.setShape(&body);
	bodyModel.glMoveTo(0.0, 5.0, 0.0);
	bodyModel.setMaterial(Yellow);
	bodyModel.setParent(&carModel);


	tireModel[0].glMoveTo(-4.0, 1.0, -8.0);
	tireModel[0].glVec(1.0, 0.0, 0.0);
	tireModel[1].glMoveTo(4.0, 1.0, -8.0);
	tireModel[1].glVec(-1.0, 0.0, 0.0);
	tireModel[2].glMoveTo(-4.0, 1.0, 8.0);
	tireModel[2].glVec(1.0, 0.0, 0.0);
	tireModel[3].glMoveTo(4.0, 1.0, 8.0);
	tireModel[3].glVec(-1.0, 0.0, 0.0);

	for(i = 0; i < 4; i++) {
		tireModel[i].setShape(&tire);
		tireModel[i].setMaterial(Gray2);
		tireModel[i].setParent(&carModel);
	}

	driverModel[0].setShape(&driver);
	driverModel[1].setShape(&driver);
	driverModel[0].glMoveTo(-2.0, 10.0, 0.0);
	driverModel[1].glMoveTo(2.0, 10.0, 0.0);
	driverModel[0].setMaterial(Cream);
	driverModel[1].setMaterial(Cream);
	driverModel[0].setSmoothMode(true);
	driverModel[1].setSmoothMode(true);
	driverModel[0].setParent(&carModel);
	driverModel[1].setParent(&carModel);

	birdModel.glMoveTo(0.0, 100.0, 200.0);
	birdModel.glFocus(0.0, 5.0, 0.0);
	birdModel.glUpvec(0.0, 1.0, 0.0);
	birdModel.setParent(&carModel);

	carModel.glMoveTo(CIRCUITX, CIRCUITY, 0.0);
	carModel.glVec(0.0, -1.0, 0.0);
	carModel.glUpvec(0.0, 0.0, 1.0);

	return;
}

void Car::entryScene(fk_Scene *scene, bool viewFlag)
{
	int		i;

	scene->entryModel(&bodyModel);

	for(i = 0; i < 4; i++) {
		scene->entryModel(&tireModel[i]);
	}

	scene->entryModel(&driverModel[0]);
	if(viewFlag == true) {
		scene->entryModel(&driverModel[1]);
	} else {
		scene->entryCamera(&driverModel[1]);
	}

	return;
}

fk_Vector Car::getCarPosition(void)
{
	return carModel.getPosition();
}

fk_Model * Car::getBirdModel(void)
{
	return &birdModel;
}

void Car::forward(void)
{
	fk_Vector	carPosition, carVelocity;
	fk_Vector	Xplus(1.0, 0.0, 0.0), Xminus(-1.0, 0.0, 0.0);
	fk_Vector	Yplus(0.0, 1.0, 0.0), Yminus(0.0, -1.0, 0.0);
	double		X = CIRCUITX;
	double		Y = CIRCUITY;

	carPosition = carModel.getPosition();
	carVelocity = carModel.getVec();
	carModel.loTranslate(0.0, 0.0, -SPEED); // Oi

	// T[Lbg̊OɂłꍇA]B
	if(carPosition.x > X) {
		rotate(&carModel, carVelocity, Xplus);
	}
	if(carPosition.x < -X) {
		rotate(&carModel, carVelocity, Xminus);
	}
	if(carPosition.y > Y) {
		rotate(&carModel, carVelocity, Yplus);
	}
	if(carPosition.y < -Y) {
		rotate(&carModel, carVelocity, Yminus);
	}

	return;
}

void Car::rotate(fk_Model *model, fk_Vector velocity, fk_Vector orgVec)
{
	// velocity  orgVec ̓ϒlA܂px 90 xȓ̏ꍇ]
	if(velocity * orgVec > EPS) {
		model->loAngle(FK_PI/100.0, 0.0, 0.0);
	}
	return;
}

void World::defLight(int lightID, double x, double y, double z)
{
	lightModel[lightID].setShape(&lightShape);
	lightModel[lightID].setMaterial(White);
	lightModel[lightID].glTranslate(0.0, 0.0, 0.0);
	lightModel[lightID].glFocus(x, y, z);

	return;
}

void World::makeBuild(int buildID, double x, double y,
					  double heightScale, fk_Material *buildMat)
{
	buildModel[buildID].setShape(&buildShape);
	buildModel[buildID].setScale(heightScale, fk_Z);
	buildModel[buildID].glMoveTo(x, y, (BUILDWIDTH * heightScale)/2.0);
	buildModel[buildID].setMaterial(*buildMat);

	return;
}

void World::init(void)
{
	// Ɩ̐ݒ
	defLight(0, 1.0, 1.0, -1.0);
	defLight(1, -1.0, -1.0, -1.0);

	// ̐ݒ
	buildShape.setSize(BUILDWIDTH, BUILDWIDTH, BUILDWIDTH);
	makeBuild(0, -250.0, 100.0, 5.0, &Red);
	makeBuild(1, -150.0, 400.0, 2.0, &DimYellow);
	makeBuild(2, 50.0, 250.0, 4.0, &Blue);
	makeBuild(3, 300.0, 200.0, 3.0, &Gray1);
	makeBuild(4, 250.0, -250.0, 0.5, &Green);
	makeBuild(5, -50.0, -350.0, 6.0, &Orange);

	// nʂ̐ݒ
	groundShape.setSize(1000, 1000, 2.0);
	groundModel.setShape(&groundShape);
	groundModel.glTranslate(0.0, 0.0, -1.0);
	groundModel.setMaterial(Brown);

	return;
}

void World::entryScene(fk_Scene *scene, bool buildFlag)
{
	int		i;

	scene->entryModel(&groundModel);

	for(i = 0; i < 2; i++) {
		scene->entryModel(&lightModel[i]);
	}

	for(i = 0; i < 6; i++) {
		scene->entryModel(&buildModel[i]);
	}

	if(buildFlag == true) scene->removeModel(&buildModel[0]);
	return;
}
